@@ -1,6 +1,7 @@
 import numpy as np
 import codecs, json
 import copy
+import random
 
 def softmax(x):
     e_x = np.exp(x - np.max(x))
@@ -34,6 +35,7 @@ class Population:
             total_fit += i.fitness
         for i in self.birds:
             odds.append(i.fitness / (total_fit - i.fitness))
+        odds /= np.sum(odds)
         print([max(odds), min(odds)])
         
         next_gen = []
@@ -52,7 +54,10 @@ class Population:
             del self.birds[-1]
         self.birds = next_gen
 
-    def mutate(self, mutate_r8, radius_ratio):
+    def mutate(self, mutate_r8, radius_ratio, immune=0, reinit=0.01):
+        police = []
+        for i in range(immune):
+            police.append(copy.deepcopy(sorted(self.birds, key=lambda x: x.fitness)[-(i + 1)]))
         for bird in self.birds:
             if np.random.uniform() < .96:
                 for matrix in range(len(bird.matrices)):
@@ -60,6 +65,13 @@ class Population:
                         for col in range(len(bird.matrices[matrix][0])):
                             if np.random.uniform() < mutate_r8:
                                 bird.matrices[matrix][row][col] += np.random.normal(0, 1/np.sqrt(len(bird.matrices[matrix][0])) * radius_ratio, 1)[0]
+                                if np.random.uniform() < reinit:
+                                    bird.matrices[matrix][row][col] = np.random.normal(0, 1/np.sqrt(len(bird.matrices[matrix][0])))[0]
+        random.shuffle(self.birds)
+        for i in range(immune):
+            del self.birds[i]
+            self.birds.append(police[i])
+        random.shuffle(self.birds)
         
 
 
